#include <stdio.h>
#include <malloc.h>
#include <string.h>

#ifndef TRUE
#define TRUE 1
#define FALSE 0
#endif

extern char *keysymtab[];

static char transtab[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789";

#define VALS (sizeof(transtab))

struct ksym { struct ksym *next; char l; char name[1]; };

static struct ksym *dbase[VALS];

static int translate(char c)
{
  int i;

  for(i=0; i<VALS; i++)
    if(transtab[i] == c)
      return i;
  return VALS-1;
}

static int checkhex(char *p)
{
  char c;

  while(c = *(p++)) {
    if(c>='0' && c<='9')
      continue;
    if(c>='A' && c<='F')
      continue;
    if(c>='a' && c<='f')
      continue;
    return FALSE;
  }
  return TRUE;
}

static void dbInit(char *fname);

void searchDatabase(char *str, int *num_compl, char **compls,
		    int *extnum, char **extstr, int *exact)
{
  int l, n, cl = 0, mn = -1;
  char *p, *q, *r, *last = (char *)0;
  static int inited = FALSE;
  struct ksym *ks;

  if(!inited) {
    for(n=0; n<VALS; n++)
      dbase[n]=(struct ksym *)0;
    inited = TRUE;
    dbInit(LIBDIR"/XKeysymDB");
  }
  *exact = FALSE;
  *num_compl = 0;
  *extnum = 0;
  if(!(l=strlen(str)))
    return;
  if(l>=3)
    if(str[0] == '0' && str[1] == 'x')
      if(checkhex(&str[2])) {
	*exact = TRUE;
	*num_compl = 1;
	if(compls) *compls = str;
	return;
      }
  for(p = keysymtab[translate(*str)]; n = *(p++); p += n)
    if( n > l ) {
      for(q = p, r = str; *r; q++, r++)
	if( *q != *r )
	  break;
      if(!*r) {
	(*num_compl) ++;
	if(compls)
	  *(compls++) = p;
	if(mn<0)
	  mn = n;
	else
	  if( n < mn )
	    mn = n;
	if(last) {
	  for(q = p, r = last; *(q++)==*(r++););
	  cl = (int)(q - p);
	} else {
	  last = p;
	  cl = n;
	}
      }
    }
  for(ks = dbase[translate(*str)]; ks; ks = ks->next)
    if( (n=ks->l) > l ) {
      for(q = ks->name, r = str; *r; q++, r++)
	if( *q != *r )
	  break;
      if(!*r) {
	(*num_compl) ++;
	if(compls)
	  *(compls++) = ks->name;
	if(mn<0)
	  mn = n;
	else
	  if( n < mn )
	    mn = n;
	if(last) {
	  for(q = ks->name, r = last; *(q++)==*(r++););
	  cl = (int)(q - ks->name);
	} else {
	  last = ks->name;
	  cl = n;
	}
      }
    }
  if(last) {
    *exact = (mn <= cl);
    *extnum = cl-l-1;
    *extstr = &last[l];
  }
}


static void dbInit(char *fname)
{
  char buf[256], name[256];
  int sym, c, e, x, n;
  char *es;
  struct ksym *new, *p;
  FILE *f;

  if(f = fopen(fname, "r")) {
    while(fgets(buf, sizeof(buf), f))
      if(*buf && *buf != '\n' && *buf != '!')
	if(sscanf(buf, "%s : %x\n", name, &sym)==2) {
	  searchDatabase(name, &c, (char **)0, &e, &es, &x);
	  if(!x)
	    if(new =
	       (struct ksym *)malloc(sizeof(struct ksym)+(c = strlen(name)))) {
	      n = translate(*name);
	      strcpy(new->name, name);
	      new->l = c+1;
	      new->next = dbase[n];
	      dbase[n] = new;
	    }
	}
    fclose(f);
    for(n=0; n<VALS; n++) {
      new = dbase[n];
      dbase[n] = NULL;
      while(new) {
	p = new->next;
	new->next = dbase[n];
	dbase[n] = new;
	new = p;
      }
    }
  }
}
