#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/Xaw/Paned.h>
#include <X11/Xaw/Viewport.h>
#include <X11/Xaw/List.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/AsciiText.h>
#include <X11/Xaw/Command.h>
#include <X11/keysym.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <dirent.h>
#include <pwd.h>
#include <stdlib.h>
#include <string.h>

#include "StringField.h"

extern char *sys_errlist[];

#ifdef __sgi
#define FORK fork /* vfork is not supported in IRIX */
#else
#define FORK vfork
#endif

typedef struct { short int len, buf; char *str; } Mstring;
typedef struct _mlistnode {
  struct _mlistnode *next;
  Mstring str;
} MListNode, *MList;

typedef struct filereq {
  void (*callback)();
  MList files;
  char **filelist;
  int filelistsize;
  Widget list, drawer, file, shell, popup, popuplabel;
  Mstring currentdir, tempstr;
} *Filereq;

#define MBLKSIZE 16
#define MBLKMASK (MBLKSIZE-1)

#define Mstr(s) ((s).str)
#define Mstrlen(s) ((s).len)
#define MLstr(n) ((n)->str)
#define MLnext(n) ((n)->next)

static void Mkill(Mstring *s)
{
  if(s->str) XtFree(s->str);
  s->str = NULL;
  s->len = s->buf = 0;
}

static void Mtrunc(Mstring *s, short int l)
{
  if(l < s->len)
    s->str[s->len = l] = '\0';
}

static void Mptrunc(Mstring *s, char *pos)
{
  if(s->str)
    Mtrunc(s, (short int)(pos - s->str));
}

static void Mstrcpy(Mstring *d, char *s)
{
  short int l = strlen(s);

  if(l >= d->buf)
    d->str = XtRealloc(d->str, d->buf = ((l+MBLKSIZE)&~MBLKMASK));
  strcpy(d->str, s);
  d->len = l;
}

static void Mstrncat(Mstring *d, char *s, short int l)
{
  if((l += d->len) >= d->buf)
    d->str = XtRealloc(d->str, d->buf = ((l+MBLKSIZE)&~MBLKMASK));
  strncpy(d->str + d->len, s, l);
  d->str[l] = '\0';
  d->len = l;
}

static void Mstrcat(Mstring *d, char *s)
{
  short int l = strlen(s) + d->len;

  if(l >= d->buf)
    d->str = XtRealloc(d->str, d->buf = ((l+MBLKSIZE)&~MBLKMASK));
  strcpy(d->str + d->len, s);
  d->len = l;
}

static MListNode *MLNremove(MListNode **head)
{
  MListNode *node;

  if(node = *head) {
    *head = node -> next;
    Mkill(&node->str);
    XtFree((char *)node);
  }
  return node;
}

static MListNode *MLNinsert(MListNode **head)
{
  MListNode *node = (MListNode *)XtCalloc(1, sizeof(MListNode));

  node->next = *head;
  return *head = node;
}

static void MLkill(MList *l)
{
  while(MLNremove(l));
}

/* ARGSUSED */
static void destroyFileReq(Widget w, XtPointer client_data, XtPointer call_data)
{
  Filereq filereq = (Filereq) client_data;

  MLkill(&filereq->files);
  if(filereq->filelist) XtFree((char *)filereq->filelist);
  Mkill(&filereq->currentdir);
  Mkill(&filereq->tempstr);
}

/* ARGSUSED */
static void popDown(Widget w, XtPointer client_data, XtPointer call_data)
{
  XtPopdown((Widget)client_data);
}

static void setCurrentDir(Filereq f)
{
  Arg arg;

  XtSetArg(arg, XtNstring, (Mstr(f->currentdir)? Mstr(f->currentdir):""));
  XtSetValues(f->drawer, &arg, 1);
  XawTextSetInsertionPoint(f->drawer, Mstrlen(f->currentdir));
}

static int qsort_comp(const void *a, const void *b)
{
  register char *s1 = *((char**)a), *s2 = *((char**)b), c, d;
  int l1, l2;

  if(l1 = strlen(s1))
    l1 = (s1[l1-1] == '/');
  if(l2 = strlen(s2))
    l2 = (s2[l2-1] == '/'); 

  if(l1 != l2)
    return (l1? -1 : 1);

  while(c = *(s1++))
    if(c != *(s2++))
      break;

  if(!c) return (*s2? -1:0);
  if(c == '/') return -1;
  if((d=s2[-1]) == '/') return 1;
  return (c < d? -1 : 1);
}

static void updateDir(Filereq filereq)
{
  DIR *dir;
  struct dirent *de;
  MListNode *new, **tail = &filereq->files;
  int ol, n = 0;
  char **p, *path;
  Widget w = filereq->list;
  struct stat st;

  if(Mstrlen(filereq->currentdir))
    path = Mstr(filereq->currentdir);
  else
    path = ".";
  if(filereq->filelist) XtFree((char *)filereq->filelist);
  filereq->filelist = NULL;
  filereq->filelistsize = 0;
  MLkill(tail);
  if(dir = opendir(path)) {
    ol = Mstrlen(filereq->currentdir);
    while(de = readdir(dir)) {
      new = MLNinsert(tail);
      tail = &MLnext(new);
      Mstrcpy(&MLstr(new), de->d_name);
      Mstrcat(&filereq->currentdir, de->d_name);
      if(Mstrlen(filereq->currentdir))
	if(stat(Mstr(filereq->currentdir), &st)>=0)
	  if((st.st_mode & S_IFMT)== S_IFDIR)
	    Mstrcat(&MLstr(new), "/");
      Mtrunc(&filereq->currentdir, ol);
      n ++;
    }
    closedir(dir);
    p = filereq->filelist = (char **)XtCalloc(n, sizeof(char *));
    filereq->filelistsize = n;
    for(new = filereq->files; new; new = MLnext(new))
      *(p++) = Mstr(MLstr(new));
    qsort((void *)filereq->filelist, n, sizeof(char *), qsort_comp);
    XawListChange(w, filereq->filelist, n, 0, True);
    XtSetSensitive(w, True);
  } else {
    Arg arg;

    n = errno;
    XtSetSensitive(w, False);
    XawListChange(w, sys_errlist + n, 1, 0, True);
    if(n == ENOENT) {
      XtSetArg(arg, XtNlabel, path);
      XtSetValues(filereq->popuplabel, &arg, 1);
      XtPopup(filereq->popup, XtGrabExclusive);
    }
  }
  XtUnmanageChild(filereq->list);
  XtManageChild(filereq->list);
}

/* ARGSUSED */
static void makeNewDir(Widget w, XtPointer client_data, XtPointer call_data)
{
  char *p;
  Arg arg;
  Filereq f = (Filereq) client_data;
  pid_t pid;

  XtSetArg(arg, XtNlabel, &p);
  XtGetValues(f->popuplabel, &arg, 1);

  if(pid = FORK()) {
    if(pid<0)
      perror("fork");
    else
      while(wait(NULL)<0)
	if(errno!=EINTR) {
	  perror("wait");
	  break;
	}
  } else {
    /* child */
    execlp("mkdir", "mkdir", "-p", p, NULL);
    perror("execlp mkdir");
    _exit(0);
  }
  updateDir(f);
  XtPopdown(f->popup);
}

static char *lastComp(char *path)
{
  char *b = path;

  while(*path)
    if(*(path++) == '/')
      if(*path) b = path;

  return (((!strcmp(b, "/"))||(!strcmp(b, "../")))? NULL : b);
}

static char *enterPath(char *path, Filereq f)
{
  char *p, *b = path;
  int dt = 0, nc = 0, c = 0;
  struct passwd *pw;

  while(*path) {
    switch(*(path++)) {
    case '.':
      dt++;
      break;
    case '/':
      if(!nc)
	switch(dt) {
	case 1:
	  break;
	case 2:
	  if(Mstrlen(f->currentdir))
	    if(p = lastComp(Mstr(f->currentdir))) {
	      Mptrunc(&f->currentdir, p);
	      break;
	    }
	default:
	  nc = dt;
	  dt = 0;
	}
      if(nc)
	if(*b == '~' && !c) {
	  if(nc>1) {
	    path[-1] = '\0';
	    if(pw = getpwnam(b+1))
	      p = pw->pw_dir;
	    else
	      p = NULL;
	    path[-1] = '/';
	  } else
	    p = getenv("HOME");
	  if(!p)
	    p = "";
	  Mstrcpy(&f->currentdir, p);
	  if(!Mstrlen(f->currentdir))
	    Mstrcat(&f->currentdir, "/");
	  else
	    if(Mstr(f->currentdir)[Mstrlen(f->currentdir)-1]!='/')
	      Mstrcat(&f->currentdir, "/");
	}
	else
	  Mstrncat(&f->currentdir, b, nc+dt+1);
      else
	if(!dt)
	  Mstrcpy(&f->currentdir, "/");
      b = path;
      dt = nc = 0;
      c ++;
      break;
    default:
      nc ++;
    }
  }
  setCurrentDir(f);
  if(c) updateDir(f);
  return b;
}

static void getCurrentDir(Widget w, XtPointer client_data, XtPointer call_data)
{
  Arg arg;
  char *p, *fm = NULL, **m;
  int n, sz, cmpl = 0, ms, c = 0;
  Filereq f = (Filereq)client_data;
  XAnyEvent *e = (XAnyEvent *)call_data;

  if(e)
    if(e->type == KeyPress)
      if(XKeycodeToKeysym(e->display, ((XKeyEvent *)e)->keycode, 0) == XK_Tab)
	cmpl ++;
  XtSetArg(arg, XtNstring, &p);
  XtGetValues(f->drawer, &arg, 1);
  Mstrcpy(&f->tempstr, p);
  if(!cmpl)
    if(Mstrlen(f->tempstr))
      if(Mstr(f->tempstr)[Mstrlen(f->tempstr)-1] != '/')
	Mstrcat(&f->tempstr, "/");
  Mstrcpy(&f->currentdir, "");
  p = enterPath(Mstr(f->tempstr),f);
  if(cmpl = strlen(p)) {
    ms = cmpl;
    for(n=0, m=f->filelist; n<f->filelistsize; n++, m++)
      if((sz = strlen(*m)-1) >= cmpl)
	if((*m)[sz]=='/')
	  if(!memcmp(p, f->filelist[n], cmpl))
	    if(c) {
	      c++;
	      sz = ms;
	      for(ms = 0; ms<sz && (*m)[ms] == fm[ms]; ms++); 
	    } else {
	      c++;
	      ms = sz+1;
	      fm = *m;
	    }
    if(c==1)
      enterPath(fm, f);
    else {
      if(ms > cmpl)
	Mstrncat(&f->tempstr, &fm[cmpl], ms-cmpl);
      else
	XBell(XtDisplay(w), 0);
      XtSetArg(arg, XtNstring, Mstr(f->tempstr));
      XtSetValues(f->drawer, &arg, 1);
      XawTextSetInsertionPoint(f->drawer, Mstrlen(f->tempstr));
    }
  }
}

/* ARGSUSED */
static void parentCB(Widget w, XtPointer client_data, XtPointer call_data)
{
  enterPath("../", (Filereq)client_data);
}

/* ARGSUSED */
static void homeCB(Widget w, XtPointer client_data, XtPointer call_data)
{
  enterPath("~/", (Filereq)client_data);
}

/* ARGSUSED */
static void rootCB(Widget w, XtPointer client_data, XtPointer call_data)
{
  enterPath("/", (Filereq)client_data);
}

static void qSelect(Filereq f, int n)
{
  Arg arg;

  XawListHighlight(f->list, n);
  Mstrcpy(&f->tempstr, f->filelist[n]);
  n = Mstrlen(f->tempstr)-1;
  if(Mstr(f->tempstr)[n] =='/')
    Mtrunc(&f->tempstr, n);
  XtSetArg(arg, XtNstring, Mstr(f->tempstr));
  XtSetValues(f->file, &arg, 1);
  XawTextSetInsertionPoint(f->file, Mstrlen(f->tempstr));
}

static void selectedFile(Widget w, XtPointer client_data, XtPointer call_data)
{
  Arg arg;
  char *op, *p;
  int n, sz, dn = 1;
  struct stat st;
  XawListReturnStruct *xlrs;
  Filereq f = (Filereq) client_data;
  XAnyEvent *e = (XAnyEvent *) call_data;

  XtSetArg(arg, XtNstring, &op);
  XtGetValues(f->file, &arg, 1);
  p = enterPath(op, f);
  if(p == op && *p == '~') {
    Mstrcpy(&f->tempstr, p);
    Mstrcat(&f->tempstr, "/");
    p = enterPath(Mstr(f->tempstr), f);
  }
  XtSetArg(arg, XtNstring, p);
  XtSetValues(f->file, &arg, 1);
  XawTextSetInsertionPoint(f->file, strlen(p));
  if(e)
    if(e->type == KeyPress)
      switch(XKeycodeToKeysym(e->display, ((XKeyEvent *)e)->keycode, 0)) {
      case XK_Up:
	dn = -1;
      case XK_Down:
	xlrs = XawListShowCurrent(f->list);
	if((n = xlrs->list_index) != XAW_LIST_NONE) {
	  Mstrcpy(&f->tempstr, xlrs->string);
	  sz = Mstrlen(f->tempstr)-1;
	  if(Mstr(f->tempstr)[sz] =='/')
	    Mtrunc(&f->tempstr, sz);
	}
	XtFree((char *)xlrs);
	if(n != XAW_LIST_NONE)
	  if(!strcmp(p, Mstr(f->tempstr))) {
	    n += dn;
	    if( n>=0 && n<f->filelistsize )
	      qSelect(f, n);
	    return;
	  }
      case XK_Tab:
	n = ((dn<0)? f->filelistsize - 1 : 0);
	if(sz = strlen(p))
	  for(; n>=0 && n<f->filelistsize; n+=dn) 
	    if(!memcmp(p, f->filelist[n], sz))
	      break;
	if( n>=0 && n<f->filelistsize )
	  qSelect(f, n);
	else
	  XBell(XtDisplay(w), 0);
	return;
      }
  if(*p) {
    n = Mstrlen(f->currentdir);
    Mstrcat(&f->currentdir, p);
    if(stat(Mstr(f->currentdir),&st)>=0)
      if((st.st_mode & S_IFMT)==S_IFDIR) {
	Mtrunc(&f->currentdir, n);
	if(!strcmp(p, "."))
	  updateDir(f);
	else if(!strcmp(p, ".."))
	  enterPath("../", f);
	else {
	  Mstrcat(&f->currentdir, p);
	  Mstrcat(&f->currentdir, "/");
	  setCurrentDir(f);
	  updateDir(f);
	}
	XtSetArg(arg, XtNstring, "");
	XtSetValues(f->file, &arg, 1);
	return;
      }
    XtPopdown(f->shell);
    f->callback(Mstr(f->currentdir));
    Mtrunc(&f->currentdir, n);
  }
}

static void listViewCB(Widget w, XtPointer client_data, XtPointer call_data)
{
  Arg arg;
  int hot, l;
  char *p;
  Filereq f = (Filereq) client_data;
  XawListReturnStruct *r = (XawListReturnStruct *) call_data;

  if(!(l = strlen(r->string)))
    return;
  XtSetArg(arg, XtNstring, &p);
  XtGetValues(f->file, &arg, 1);
  hot = (r->string[l-1]=='/' || !strcmp(p, r->string));
  XtSetArg(arg, XtNstring, r->string);
  XtSetValues(f->file, &arg, 1);
  XawTextSetInsertionPoint(f->file, l);
  if(hot)
    selectedFile(w, (XtPointer)f, NULL);
}

static char filetrans[] =
"<Key>Tab: notify()\n\
<Key>Up: notify()\n\
<Key>Down: notify()";

static char drwtrans[] =
"<Key>Tab: notify()";

static char fileaccel[] =
"<Key>Linefeed: notify() \n\
<Key>Return: notify() \n\
<Key>KP_Enter: notify() \n\
<Key>Tab: notify()\n\
<Key>Up: notify()\n\
<Key>Down: notify()";

static char canaccel[] =
"<Key>Escape: set() notify() unset()";

Widget createFileRequester(Widget parent, char *name, char *title, void (*callback)())
{
  Widget paned, form, files, label1, label2;
  Widget ok, cancel, parentdir, home, root, popupok;
  Filereq filereq = (Filereq)XtCalloc(1, sizeof(struct filereq));

  filereq -> callback = callback;

  filereq -> shell =
    XtVaCreatePopupShell(name, transientShellWidgetClass, parent,
			 XtNtitle, title, NULL);

  XtAddCallback(filereq->shell, XtNdestroyCallback, destroyFileReq, (XtPointer)filereq);

  paned = XtVaCreateManagedWidget("request", panedWidgetClass, filereq->shell,
				 NULL);
  form = XtVaCreateManagedWidget("listview", formWidgetClass, paned, NULL);
  files = XtVaCreateManagedWidget("files", viewportWidgetClass, form,
				  XtNallowVert, TRUE, XtNforceBars, TRUE,
				  XtNuseRight, TRUE,
				  XtNleft, XtChainLeft, XtNright, XtChainRight,
				  XtNtop, XtChainTop, XtNbottom, XtChainBottom,
				  NULL);

  filereq -> list =
    XtVaCreateManagedWidget("list", listWidgetClass, files,
			    XtNdefaultColumns, 1, XtNforceColumns, TRUE,
			    NULL);

  form = XtVaCreateManagedWidget("stringinput", formWidgetClass, paned,
				 XtNskipAdjust, TRUE, XtNshowGrip, FALSE,
				 NULL);
  label1 = XtVaCreateManagedWidget("drawerlabel", labelWidgetClass, form,
				   XtNlabel, "Drawer:", XtNborderWidth, 0,
				   XtNleft, XtChainLeft, XtNright, XtChainLeft,
				   XtNtop, XtChainTop, XtNbottom, XtChainTop,
				   NULL);
  filereq -> drawer =
    XtVaCreateManagedWidget("drawer", stringFieldWidgetClass, form,
			    XtNeditType, XawtextEdit,
			    XtNfromHoriz, label1,
			    XtNleft, XtChainLeft, XtNright, XtChainRight,
			    XtNtop, XtChainTop, XtNbottom, XtChainTop,
			    NULL);
  label2 = XtVaCreateManagedWidget("filelabel", labelWidgetClass, form,
				   XtNlabel, "File:", XtNborderWidth, 0,
				   XtNfromVert, label1,
				   XtNleft, XtChainLeft, XtNright, XtChainLeft,
				   XtNtop, XtChainBottom, XtNbottom, XtChainBottom,
				   NULL);
  filereq -> file =
    XtVaCreateManagedWidget("file", stringFieldWidgetClass, form,
			    XtNeditType, XawtextEdit,
			    XtNaccelerators,
			    XtParseAcceleratorTable(fileaccel),
			    XtNfromHoriz, label1, XtNfromVert, label1,
			    XtNleft, XtChainLeft, XtNright, XtChainRight,
			    XtNtop, XtChainBottom, XtNbottom, XtChainBottom,
			    NULL);

  form = XtVaCreateManagedWidget("buttons", formWidgetClass, paned,
				 XtNskipAdjust, TRUE, XtNshowGrip, FALSE,
				 NULL);
  ok = XtVaCreateManagedWidget("ok", commandWidgetClass, form,
			       XtNlabel, "Ok",
			       XtNtop, XtChainTop, XtNbottom, XtChainBottom,
			       XtNleft, XtChainLeft,
			       NULL);
  parentdir = XtVaCreateManagedWidget("parent", commandWidgetClass, form,
				      XtNlabel, "Parent", XtNfromHoriz, ok,
				      XtNtop, XtChainTop, XtNbottom, XtChainBottom,
				      NULL);
  home = XtVaCreateManagedWidget("home", commandWidgetClass, form,
				 XtNlabel, "Home", XtNfromHoriz, parentdir,
				 XtNtop, XtChainTop, XtNbottom, XtChainBottom,
				 NULL);
  root = XtVaCreateManagedWidget("root", commandWidgetClass, form,
				 XtNlabel, "Root", XtNfromHoriz, home,
				 XtNtop, XtChainTop, XtNbottom, XtChainBottom,
				 NULL);
  cancel = XtVaCreateManagedWidget("cancel", commandWidgetClass, form,
				   XtNaccelerators,
				   XtParseAcceleratorTable(canaccel),
				   XtNlabel, "Cancel", XtNfromHoriz, root,
				   XtNtop, XtChainTop, XtNbottom, XtChainBottom,
				   XtNright, XtChainRight,
				   NULL);


  filereq->popup =
    XtVaCreatePopupShell("createdirpopup", transientShellWidgetClass,
			 filereq->shell, XtNallowShellResize, TRUE,
			 XtNtitle, "Request", NULL);

  form =
    XtVaCreateManagedWidget("form", formWidgetClass, filereq->popup, NULL);

  label1=XtVaCreateManagedWidget("text1", labelWidgetClass, form,
				 XtNborderWidth, 0,
				 XtNleft, XtChainLeft, XtNright, XtChainRight,
				 XtNtop, XtChainTop,
				 XtNlabel, "The directory", NULL);
  filereq->popuplabel =
    XtVaCreateManagedWidget("text2", labelWidgetClass, form,
			    XtNborderWidth, 0,
			    XtNfromVert, label1,
			    XtNleft, XtChainLeft, XtNright, XtChainRight,
			    XtNresizable, TRUE,
			    NULL);

  label2=XtVaCreateManagedWidget("text3", labelWidgetClass, form,
				 XtNborderWidth, 0,
				 XtNlabel, "does not exist.  Create it?",
				 XtNfromVert, filereq->popuplabel,
				 XtNleft, XtChainLeft, XtNright, XtChainRight,
				 NULL);

  popupok = XtVaCreateManagedWidget("ok", commandWidgetClass, form,
				    XtNlabel, "Ok",
				    XtNbottom, XtChainBottom,
				    XtNleft, XtChainLeft,
				    XtNfromVert, label2,
				    NULL);

  XtAddCallback(XtVaCreateManagedWidget("cancel", commandWidgetClass, form,
					XtNlabel, "Cancel",
					XtNbottom, XtChainBottom,
					XtNfromHoriz, popupok,
					XtNfromVert, label2,
					NULL),
		XtNcallback, popDown, (XtPointer)filereq->popup);

  XtOverrideTranslations(filereq->drawer, XtParseTranslationTable(drwtrans));
  XtOverrideTranslations(filereq->file, XtParseTranslationTable(filetrans));
  XtInstallAccelerators(paned, filereq->file);
  XtInstallAccelerators(paned, cancel);

  XtAddCallback(filereq->list, XtNcallback, listViewCB, (XtPointer)filereq);

  XtAddCallback(filereq->drawer, XtNcallback, getCurrentDir, (XtPointer)filereq);
  XtAddCallback(filereq->file, XtNcallback, selectedFile, (XtPointer)filereq);

  XtAddCallback(ok, XtNcallback, selectedFile, (XtPointer)filereq);
  XtAddCallback(parentdir, XtNcallback, parentCB, (XtPointer)filereq);
  XtAddCallback(home, XtNcallback, homeCB, (XtPointer)filereq);
  XtAddCallback(root, XtNcallback, rootCB, (XtPointer)filereq);
  XtAddCallback(cancel, XtNcallback, popDown, (XtPointer)filereq->shell);

  XtAddCallback(popupok, XtNcallback, makeNewDir, (XtPointer)filereq);

  updateDir(filereq);

  return filereq->shell;
}

