#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Paned.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Dialog.h>
#include <X11/Xaw/SimpleMenu.h>
#include <X11/Xaw/SmeBSB.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <dirent.h>
#include <sys/stat.h>
#include "xsetmap.h"

static Widget tempkey, layout, loadkbd, savekbd;
static Widget alert_shell, alert_dialog;
static int tempkeycode;
static int key_x = 0, key_y = 0, key_w, key_h;

static char homedir[512];

static char *kbdpaths[] = {
  homedir,
  LIBDIR"/xsetmap",
  "/usr/local/share/xsetmap",
  NULL
};

/*VARARGS1*/
void alert(char *fmt, ...)
{
  char buf[512];
  Arg arg[3];
  va_list va;
  /*CONSTCOND*/
  va_start(va, fmt);
  vsprintf(buf, fmt, va);
  XtSetArg(arg[0], XtNlabel, buf);
  XtSetValues(alert_dialog, arg, 1);
  va_end(va);
  XtPopup(alert_shell, XtGrabNone);
}

static void setTempKey(char *label, int code)
{
  Arg arg[1];

  XtSetArg(arg[0], XtNlabel, label);
  XtSetValues(tempkey, arg, 1);
  tempkeycode = code;
}

static void setTempKey0()
{
  setTempKey("?", -1);
}

static void setKeySize(int w, int h)
{
  Arg arg[2];

  XtSetArg(arg[0], XtNwidth, w*appldata.blocksize-1);
  XtSetArg(arg[1], XtNheight, h*appldata.blocksize-1);
  XtSetValues(tempkey, arg, 2);
  key_w = w;
  key_h = h;
}

/* ARGSUSED */
void popdownCallback(Widget w, Widget shell)
{
  XtPopdown(shell);
}

/* ARGSUSED */
void popupCallback(Widget w, Widget shell)
{
  XtPopup(shell, XtGrabNone);
}

int coordEvent(XEvent *event)
{
  switch(event->type) {
  case ButtonPress:
  case ButtonRelease:
  case KeyPress:
  case KeyRelease:
  case MotionNotify:
    return TRUE;
  }
  return FALSE;
}

/*ARGSUSED*/
static void moveKey(Widget w, XEvent *event)
{
  int blocksize = appldata.blocksize;

  if(coordEvent(event)) {
    key_x = ((XMotionEvent *)event)->x / blocksize;
    if(key_x < 0) key_x = 0;
    if(key_x > 255) key_x = 255;
    key_y = ((XMotionEvent *)event)->y / blocksize;
    if(key_y < 0) key_y = 0;
    if(key_y > 255) key_y = 255;
    XtMoveWidget(tempkey, key_x * blocksize, key_y * blocksize);
  }
}

/* ARGSUSED */
static void mapKey(Widget w, XEvent *event)
{
  if(event->type == EnterNotify)
    if(((XCrossingEvent*)event)->detail == NotifyInferior)
      return;
  XtMapWidget(tempkey);
}

/* ARGSUSED */
static void unmapKey(Widget w, XEvent *event)
{
  if(event->type == LeaveNotify)
    if(((XCrossingEvent *)event)->detail == NotifyInferior)
      return;
  XtUnmapWidget(tempkey);
}

static int find_key(int x, int y, int w, int h)
{
  int a,b,c;
  struct key *k;

  for(c = 0, k = key; c < 256; c++, k++)
    if(k->w)
      for(a = 0; a < w; a++)
	for(b = 0; b < h; b++)
	  if(x+a >= k->x && y+b >= k->y &&
	     x+a < k->x+k->w && y+b < k->y+k->h)
	    return c;
  return -1;
}

static void createKeyWidget(struct key *k)
{
  char name[20],label[16];
  int blocksize = appldata.blocksize;

  sprintf(label, "$%02x", k->keycode);
  sprintf(name,  "key_%02x", k->keycode);
  if(k->widget2)
    XtDestroyWidget(k->widget2);
  k->widget2 = XtVaCreateManagedWidget(name, labelWidgetClass, layout,
				       XtNx, k->x * blocksize,
				       XtNy, k->y * blocksize,
				       XtNwidth, k->w * blocksize - 1,
				       XtNheight, k->h * blocksize - 1,
				       XtNlabel, label,
				       NULL);
  if(XtIsRealized(k->widget2))
    XLowerWindow(XtDisplay(k->widget2), XtWindow(k->widget2));
}

static void add_key(KeyCode c)
{
  struct key *k;

  if(find_key(key_x, key_y, key_w, key_h)>=0)
    return;
  k = &key[c];
  k->x = key_x;
  k->y = key_y;
  k->w = key_w;
  k->h = key_h;
  createKeyWidget(k);
  setTempKey0();
}

static void placeKey()
{
  if(tempkeycode >= 0)
    add_key(tempkeycode);
}

/* ARGSUSED */
static void newKey(Widget w, XEvent *event)
{
  if(event->type == KeyPress || event->type == KeyRelease)
    add_key(((XKeyEvent *)event)->keycode);
}

static void pickKey()
{
  char label[16];
  struct key *k;
  int c;

  c = find_key(key_x, key_y, 1, 1);
  if(c<0)
    return;
  k = &key[c];
  sprintf(label, "$%02x", c);
  setTempKey(label, c);
  setKeySize(k->w, k->h);
  k->w = 0;
  if(k->widget2) {
    XtDestroyWidget(k->widget2);
    k->widget2 = NULL;
  }
}

static void beginChangeSize()
{
  int blocksize = appldata.blocksize;

  XWarpPointer(XtDisplay(layout), None, None, 0, 0, 0, 0,
	       (key_w - 1) * blocksize, (key_h - 1) * blocksize);
}

static void endChangeSize()
{
  int blocksize = appldata.blocksize;

  XWarpPointer(XtDisplay(layout), None, None, 0, 0, 0, 0,
	       (-key_w + 1) * blocksize, (-key_h + 1) * blocksize);
}

/*ARGSUSED*/
static void changeSize(Widget w, XEvent *event)
{
  int blocksize = appldata.blocksize;
  int dx,dy;

  if(coordEvent(event)) {
    dx = ((XMotionEvent *)event)->x / blocksize - key_x + 1;
    dy = ((XMotionEvent *)event)->y / blocksize - key_y + 1;
    if(dx < 4) dx = 4;
    if(dx > 255) dx = 255;
    if(dy < 4) dy = 4;
    if(dy > 255) dy = 255;
    setKeySize(dx, dy);
  }
}

static struct k_header { char hdr[7]; char ver; } header = {
  { 'X', 'S', 'K', 'K', 'C', 'T', 'F' },
  1
};

static int checkheader(FILE *f)
{
  struct k_header khdr;

  fread(&khdr, sizeof(khdr), 1, f);
  return memcmp(&khdr, &header, sizeof(khdr));
}

int loadKeyboard(char *fname)
{
  FILE *f;
  int i,a,b;
  char buf[512], *full, *fn, **path = kbdpaths;

  fn = ((full = strchr(fname, '/'))? fname : buf);
  while(*path) {
    if(!full) {
      strcpy(buf, *(path++));
      strcat(buf, "/");
      strcat(buf, fname);
    }
    if(f = fopen(fn, "rb")) {
      if(checkheader(f)) {
	alert("\"%s\" is not a keyboard layout file", fn);
	fclose(f);
	return FALSE;
      }
      clearKeys();
      a=fgetc(f) & 0xff;
      b=fgetc(f) & 0xff;
      for(i=a; i<=b; i++) {
	key[i].x = fgetc(f);
	key[i].y = fgetc(f);
	key[i].w = fgetc(f);
	key[i].h = fgetc(f);
	if(key[i].w)
	  createKeyWidget(&key[i]);
      }
      fclose(f);
      return TRUE;
    } else
      if(full) break;
  }
  alert("Unable to open file \"%s\"", fn);
  return FALSE;
}

static void saveKbd(char *fname)
{
  FILE *f;
  int i, a, b;
  struct key *k;

  if(f = fopen(fname, "wb")) {
    fwrite(&header, sizeof(header), 1, f);
    for(a=0; a<256; a++)
      if(key[a].w) break;
    for(b=255; b>=0; --b)
      if(key[b].w) break;
    if(a>b) {
      a = 0xff;
      b = 0;
    }
    fputc(a, f);
    fputc(b, f);
    for(i = a, k = &key[a]; i<=b; i++, k++) {
      fputc(k->x, f);
      fputc(k->y, f);
      fputc(k->w, f);
      fputc(k->h, f);
    }
    if(fclose(f)<0)
      alert("Error writing file \"%s\"", fname);
  } else alert("Unable to create file \"%s\"", fname);
}

static void menuKbd(Widget w)
{
  if(loadKeyboard(XtName(w)))
    applyLayout();
}

static Widget findchild(Widget w, char *name)
{
  WidgetList l;
  Widget c;
  Arg arg[2];
  int n;

  XtSetArg(arg[0], XtNchildren, &l);
  XtSetArg(arg[1], XtNnumChildren, &n);
  XtGetValues(w, arg, 2);
  while(n--)
    if(!strcmp(name, XtName(c = *(l++))))
      return c;
  return (Widget)NULL;
}

static void addKeyboards(Widget w, char *path)
{
  DIR *d;
  struct dirent *de;
  struct stat s;
  char fn[512], *p;
  FILE *f;

  strcpy(fn, path);
  p = &fn[strlen(fn)];
  *(p++) = '/';
  if(d = opendir(path)) {
    while(de = readdir(d)) {
      strcpy(p, de->d_name);
      if(!stat(fn, &s))
	if((s.st_mode & S_IFMT) == S_IFREG)
	  if(f = fopen(fn, "rb")) {
	    if(!checkheader(f))
	      if(!findchild(w, de->d_name))
		XtAddCallback(XtVaCreateManagedWidget(de->d_name,
						      smeBSBObjectClass,
						      w, NULL),
			      XtNcallback, (XtCallbackProc)menuKbd, NULL);
	    fclose(f);
	  }
    }
    closedir(d);
  }
}

int makeKbdMenu(Widget w)
{
  Widget menu;
  Arg arg;
  Cardinal n;
  char *home, **path;

  if(home = getenv("HOME"))
    strcpy(homedir, home);
  else
    strcpy(homedir, ".");
  strcat(homedir, "/.xsetmap");
  menu = XtVaCreatePopupShell("menu", simpleMenuWidgetClass, w, NULL);
  for(path = kbdpaths; *path; path++)
    addKeyboards(menu, *path);
  XtSetArg(arg, XtNnumChildren, &n);
  XtGetValues(menu, &arg, 1);
  return n;
}

static XtActionsRec actions[] = {
  {"movekey", (XtActionProc)moveKey},
  {"mapkey", (XtActionProc)mapKey},
  {"unmapkey", (XtActionProc)unmapKey},
  {"pickkey", (XtActionProc)pickKey},
  {"placekey", (XtActionProc)placeKey},
  {"newkey", (XtActionProc)newKey},
  {"changesize", (XtActionProc)changeSize},
  {"warpdown", (XtActionProc)beginChangeSize},
  {"warpup", (XtActionProc)endChangeSize},
  {"deletekey", (XtActionProc)setTempKey0},
};

static char layouttrans[] = "#override\n\
Button3<PtrMoved>: changesize()\n\
<Btn3Down>: warpdown()\n\
<Btn3Up>: warpup()\n\
<PtrMoved>: movekey()\n\
~Button3<Enter>: mapkey()\n\
~Button3<Leave>: unmapkey()\n\
<Btn1Down>: pickkey()\n\
Button3<Btn1Up>: deletekey()\n\
<Btn1Up>: placekey()\n\
<KeyPress>: newkey()\n\
<Btn2Down>: pickkey() deletekey()";

static void createAlert(Widget *shell, Widget *dialog, Widget parent)
{
  *shell = XtVaCreatePopupShell("alert", transientShellWidgetClass, parent,
				XtNtitle, "Error!", NULL);
  *dialog = XtVaCreateManagedWidget("request", dialogWidgetClass, *shell, NULL);
  XtAddCallback(XtVaCreateManagedWidget("ok", commandWidgetClass, *dialog,
					NULL),
		XtNcallback, (XtCallbackProc)popdownCallback, (XtPointer)*shell);
}

Widget createKeyCodeWin(Widget parent)
{
  Widget shell, form, commands;

  XtAppAddActions(ctx, actions, XtNumber(actions));
  shell = XtVaCreatePopupShell("keycodes", topLevelShellWidgetClass, parent,
			       NULL);
  loadkbd = createFileRequester(shell, "loadfile",
				"Select keyboard file to load", (void(*)())loadKeyboard);
  savekbd = createFileRequester(shell, "savefile", 
				"Select keyboard file to save", saveKbd);
  createAlert(&alert_shell, &alert_dialog, shell);
  form = XtVaCreateManagedWidget("panes", panedWidgetClass, shell, NULL);
  commands = XtVaCreateManagedWidget("commands", boxWidgetClass, form,
				     XtNorientation, XtorientHorizontal,
				     XtNresizeToPreferred, TRUE,
				     NULL);
  layout = XtVaCreateManagedWidget("layout", compositeWidgetClass, form,
				   XtNshowGrip, FALSE,
				   NULL);
  XtAddCallback(XtVaCreateManagedWidget("apply", commandWidgetClass, commands,
					NULL),
		XtNcallback, (XtCallbackProc)applyLayout, NULL);
  XtAddCallback(XtVaCreateManagedWidget("load", commandWidgetClass, commands,
					NULL),
		XtNcallback, (XtCallbackProc)popupCallback, (XtPointer)loadkbd);
  XtAddCallback(XtVaCreateManagedWidget("save", commandWidgetClass, commands,
					NULL),
		XtNcallback, (XtCallbackProc)popupCallback, (XtPointer)savekbd);
  XtAddCallback(XtVaCreateManagedWidget("done", commandWidgetClass, commands,
					NULL),
		XtNcallback, (XtCallbackProc)popdownCallback, (XtPointer)shell);

  XtOverrideTranslations(layout, XtParseTranslationTable(layouttrans));

  tempkey = XtVaCreateWidget("tempkey", labelWidgetClass, layout,
			     XtNmappedWhenManaged, FALSE, XtNsensitive, FALSE,
			     XtNresize, FALSE, NULL);
  setTempKey0();
  setKeySize(4,4);
  return shell;
}
