#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Paned.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/MenuButton.h>
#include <X11/Xaw/Toggle.h>
#include <X11/Xaw/List.h>
#include <X11/Xaw/SimpleMenu.h>
#include <X11/Xaw/SmeBSB.h>
#include <X11/Xaw/AsciiText.h>
#include <X11/Xaw/Dialog.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>

#include "xsetmap.h"

#ifdef __sgi
#define FORK fork /* vfork is not supported in IRIX */
#else
#define FORK vfork
#endif

static Widget kbdwidget, symwidget = NULL, symbox = NULL, modbox = NULL,
  keysymname, parent, readreq, writereq, modwidget[8];

static struct key *currkey = NULL;

static XtTranslations keytrans;

static char *mod_name[] = { "shift", "lock", "control", "mod1",
			      "mod2", "mod3", "mod4", "mod5" };

/*ARGSUSED*/
static void changeMod(Widget w, XtPointer mp)
{
  XModifierKeymap *xk;
  int m = (int)mp;

  if(currkey) {
    currkey->modifiers ^= (1<<m);
    xk = XGetModifierMapping(XtDisplay(topLevel));
    xk = (((currkey->modifiers)&(1<<m))?
	  XInsertModifiermapEntry(xk, currkey->keycode, m)
	  :XDeleteModifiermapEntry(xk, currkey->keycode, m));
    XSetModifierMapping(XtDisplay(topLevel), xk);
    XFreeModifiermap(xk);
  }
}

static void selectKc(Widget w, struct key *k)
{
  static char buf[256*MAXKEYSYM], *b = buf;
  static char *list[MAXKEYSYM+1];
  static char *dummy = NULL;
  char *p;
  int i,m;
  static Widget highlighted = NULL;
  Arg arg;

  if(symwidget) XawListChange(symwidget, &dummy, 0, 0, FALSE);
  if(highlighted) {
    XtSetArg(arg, XtNstate, FALSE);
    XtSetValues(highlighted, &arg, 1);
  }
  if(highlighted = w) {
    XtSetArg(arg, XtNstate, TRUE);
    XtSetValues(highlighted, &arg, 1);
  }
  if(currkey = k) {
    if(symbox) XtSetSensitive(symbox, TRUE);
    if(modbox) XtSetSensitive(modbox, TRUE);
  } else {
    if(symbox) XtSetSensitive(symbox, FALSE);
    if(modbox) {
      XtSetSensitive(modbox, FALSE);
      XtSetArg(arg, XtNstate, FALSE);
      for(i=0; i<8; i++)
	XtSetValues(modwidget[i], &arg, 1);
    }
    return;
  }
  for(m=MAXKEYSYM; m; --m)
    if(k->keysym[m-1] != NoSymbol)
      break;
  for(i=0; i<m; i++) {
    if(p = XKeysymToString(k->keysym[i]))
      strcpy(b, p);
    else
      sprintf(b, "0x%04lx", (long)k->keysym[i]);
    list[i] = b;
    b+=strlen(b)+1;
  }
  list[i] = NULL;
  if(symwidget)
    XawListChange(symwidget, list, m, 0, FALSE);
  if(modbox)
    for(i=0; i<8; i++) {
      XtSetArg(arg, XtNstate, ((k->modifiers & (1<<i))? TRUE:FALSE));
      XtSetValues(modwidget[i], &arg, 1);
    }
}

static void refreshKc()
{
  if(currkey)
    selectKc(currkey->widget1, currkey);
}

static void modifyKeySyms(struct key *k)
{
  int num;

  for(num = MAXKEYSYM; num; --num)
    if(currkey->keysym[num-1] != NoSymbol)
      break;
  if(!num) num++;
  XChangeKeyboardMapping(XtDisplay(topLevel), k->keycode,
			 num, k->keysym, 1);
}

extern void searchDatabase(char *, int *, char **, int *, char **, int *);

static KeySym searchExpand(int *num_compl, char **compls, int *num_expand)
{
  char *p, *q;
  int n, exact;
  Arg arg;
  XawTextBlock blk;
  XawTextPosition pos;

  XtSetArg(arg, XtNstring, &p);
  XtGetValues(keysymname, &arg, 1);
  searchDatabase(p, num_compl, compls, &n, &q, &exact);
  if(n) {
    pos = strlen(p);
    blk.firstPos = 0;
    blk.length = n;
    blk.ptr = q;
    blk.format = FMT8BIT;
    XawTextReplace(keysymname, pos, pos, &blk);
    XawTextSetInsertionPoint(keysymname, pos + n);
    XtGetValues(keysymname, &arg, 1);
  }
  XawTextSetInsertionPoint(keysymname, strlen(p));
  *num_expand = n;
  if(exact)
    if(sscanf(p, "0x%x", &n))
      return (KeySym)n;
    else
      return XStringToKeysym(p);
  else
    return NoSymbol;
}

static void removeParent(Widget w)
{
  Widget p = XtParent(w);

  XtPopdown(p);
  XtDestroyWidget(p);
}

static void setkeysymName(char *n)
{
  Arg arg;

  XtSetArg(arg, XtNstring, n);
  XtSetValues(keysymname, &arg, 1);
  XawTextSetInsertionPoint(keysymname, strlen(n));
}

static void menuCSelect(Widget w, XtPointer callback, XawListReturnStruct *l)
{
  setkeysymName(l->string);
  if(callback) ((void(*)())(callback))();
  removeParent(w);
}

/*ARGSUSED*/
static void menuCDone(Widget w, char *data)
{
  XtFree(data);
}

static int complx=100, comply=100;

static KeySym menuComplete(int n, char **c, void (*callback)())
{
  Widget menu, list;
  int sw, sh, w, h, x = complx, y = comply;
  Arg arg[2];
  static char compltrans[] = "#augment\n<BtnDown>: removeparent()";

  sw = WidthOfScreen(XtScreen(keysymname));
  sh = HeightOfScreen(XtScreen(keysymname));
  
  menu = XtVaCreatePopupShell("completitions", overrideShellWidgetClass,
			      keysymname, NULL);
  list = XtVaCreateManagedWidget("list", listWidgetClass, menu,
				 XtNlist, c, XtNnumberStrings, n,
				 XtNverticalList, TRUE,
				 NULL);
  XtSetArg(arg[0], XtNwidth, &w);
  XtSetArg(arg[1], XtNheight, &h);
  XtGetValues(list, arg, 2);
  if(x + w > sw) x = sw - w;
  if(y + h > sh) y = sh - h;
  if(x < 0) x = 0;
  if(y < 0) y = 0;
  XtSetArg(arg[0], XtNx, x);
  XtSetArg(arg[1], XtNy, y);
  XtSetValues(menu, arg, 2);
  XtAddCallback(list, XtNcallback,
		(XtCallbackProc)menuCSelect, (XtPointer)callback);
  XtAddCallback(list, XtNdestroyCallback,
		(XtCallbackProc)menuCDone, (XtPointer)c);
  XtAugmentTranslations(list, XtParseTranslationTable(compltrans));
  XtPopup(menu, XtGrabExclusive);
  return NoSymbol;
}

/*ARGSUSED*/
static void placeCompl(Widget w, XEvent *e)
{
  if(coordEvent(e)) {
    complx = ((XButtonEvent *)e)->x_root - 2;
    comply = ((XButtonEvent *)e)->y_root - 2;
  }
}

static KeySym getKeySym(void (*callback)())
{
  KeySym ks;
  char **c;
  int n, ne;

  if((ks = searchExpand(&n, NULL, &ne)) != NoSymbol)
    if(callback)
      return ks;
  if(!n)
    XBell(XtDisplay(topLevel), 0);
  if(callback || !ne)
    if(n>1) {
      c = (char **)XtCalloc(n, sizeof(char *));
      searchExpand(&n, c, &ne);
      ks = menuComplete(n, c, callback);
    }
  return ks;
}

/*ARGSUSED*/
static void editsymCallback(Widget w, XtPointer dummy, XawListReturnStruct *xlrs)
{
  setkeysymName(xlrs->string);
}

static void replaceKeySym()
{
  XawListReturnStruct *xlrs = XawListShowCurrent(symwidget);
  int ix = xlrs->list_index;
  KeySym ks;

  XtFree((char *)xlrs);
  if(currkey)
    if((ks = getKeySym(replaceKeySym)) != NoSymbol) {
      if(ix < 0)
	for(ix = MAXKEYSYM; ix; --ix)
	  if(currkey->keysym[ix-1] != NoSymbol)
	    break;
      if(ix >= MAXKEYSYM) ix = MAXKEYSYM - 1;
      currkey->keysym[ix] = ks;
      modifyKeySyms(currkey);
    }
}

static void insertsymCallback()
{
  XawListReturnStruct *xlrs = XawListShowCurrent(symwidget);
  int i, ix = xlrs->list_index;
  KeySym ks;

  XtFree((char *)xlrs);
  if(currkey) {
    if((ks = getKeySym(insertsymCallback)) == NoSymbol)
      return;
    if(ix < 0)
      for(ix = MAXKEYSYM; ix; --ix)
	if(currkey->keysym[ix-1] != NoSymbol)
	  break;
    if(ix >= MAXKEYSYM) ix = MAXKEYSYM - 1;
    for(i = MAXKEYSYM - 1; i > ix; --i)
      currkey->keysym[i] = currkey->keysym[i-1];
    currkey->keysym[ix] = ks;
    modifyKeySyms(currkey);
  }
}

static void removesymCallback()
{
  XawListReturnStruct *xlrs = XawListShowCurrent(symwidget);
  int ix = xlrs->list_index;

  XtFree((char *)xlrs);
  if(currkey) {
    if(ix < 0)
      return;
    while(++ix < MAXKEYSYM)
      currkey->keysym[ix-1] = currkey->keysym[ix];
    currkey->keysym[MAXKEYSYM-1] = NoSymbol;
    modifyKeySyms(currkey);
  }
}

static void clearsymCallback()
{
  int i;

  if(currkey) {
    for(i = 0; i < MAXKEYSYM; i++)
      currkey->keysym[i] = NoSymbol;
    modifyKeySyms(currkey);
  }
}

/*ARGSUSED*/
static void alphaCallback(Widget w, XtPointer n)
{
  int i;

  if(currkey) {
    for(i = 1; i < MAXKEYSYM; i++)
      currkey->keysym[i] = NoSymbol;
    currkey->keysym[0] = (KeySym)n;
    modifyKeySyms(currkey);
  }
}

static void quitOut()
{
  exit(0);
}

void applyLayout()
{
  int i, blocksize = appldata.blocksize, x1 = -1, y1 = -1, x2 = -1, y2 = -1;
  struct key *k;
  Arg arg;

  selectKc(NULL, NULL);
  XawFormDoLayout(kbdwidget, False);
  for(i = 0, k = key; i<256; i++, k++) {
    if(k->widget1)
      XtDestroyWidget(k->widget1);
    if(k->w)
      if(x1>=0) {
	if(k->x < x1) x1 = k->x;
	if(k->y < y1) y1 = k->y;
	if(k->x+k->w > x2) x2 = k->x+k->w;
	if(k->y+k->h > y2) y2 = k->y+k->h;
      }
      else {
	x2 = (x1 = k->x) + k->w;
	y2 = (y1 = k->y) + k->h;
      }
  }
  if(x1>=0) {
    XtSetArg(arg, XtNwidth, (x1+x2)*blocksize);
    XtSetValues(parent, &arg, 1);
    XtSetArg(arg, XtNheight, (y1+y2)*blocksize);
    XtSetValues(kbdwidget, &arg, 1);
  }
  for(i = 0, k = key; i<256; i++, k++) {
    if(k->w) {
      char name[20], *label;

      sprintf(name,  "key_%02x", k->keycode);
      label = XKeysymToString(k->keysym[0]);
      k->widget1 = XtVaCreateManagedWidget(name, toggleWidgetClass, kbdwidget,
					   XtNhorizDistance, k->x * blocksize,
					   XtNvertDistance, k->y * blocksize,
					   XtNwidth, k->w * blocksize - 1,
					   XtNheight, k->h * blocksize - 1,
					   XtNlabel, (label? label : ""),
					   NULL);
      XtAddCallback(k->widget1, XtNcallback, (XtCallbackProc)selectKc, (XtPointer)k);
      XtOverrideTranslations(k->widget1, keytrans);
    }
    else k->widget1 = NULL;
  }
  XawFormDoLayout(kbdwidget, True);
}

void updateKeysyms(int start, int end)
{
  KeySym *ksym;
  int kpk, min, max, i, j, k;
  char *label;
  Arg arg;

  if(start >= end) return;
  XDisplayKeycodes(XtDisplay(topLevel), &min, &max);
  if(min < start) min = start;
  if(max >= end) max = end-1;
  ksym = XGetKeyboardMapping(XtDisplay(topLevel), min, max-min+1, &kpk);
  if(kpk > MAXKEYSYM) {
    k = MAXKEYSYM;
    fprintf(stderr, "XSetMap: Too many keysyms per keycode\n");
  } else k = kpk;
  for(i=start; i<end; i++) {
    if(i<min || i>max)
      for(j=0; j<MAXKEYSYM; j++)
	key[i].keysym[j] = NoSymbol;
    else {
      for(j=0; j<k; j++)
	key[i].keysym[j] = ksym[(i-min)*kpk+j];
      while(j<MAXKEYSYM)
	key[i].keysym[j++] = NoSymbol;
    }
    if(key[i].widget1) {
      label = XKeysymToString(key[i].keysym[0]);
      XtSetArg(arg, XtNlabel, (label? label : ""));
      XtSetValues(key[i].widget1, &arg, 1);
    }
  }
  XFree((char *)ksym);
  refreshKc();
}

void updateModifiers()
{
  int i,j;
  XModifierKeymap *mk = XGetModifierMapping(XtDisplay(topLevel));
  KeyCode *p = mk -> modifiermap;

  for(i=0; i<256; i++)
    key[i].modifiers = 0;
  for(i=0; i<8; i++)
    for(j=0; j<mk->max_keypermod; j++, p++)
      if(*p)
	key[*p].modifiers |= 1<<i;
  XFreeModifiermap(mk);
  refreshKc();
}

/*ARGSUSED*/
static void remapKbd(Widget w, XEvent *event)
{
  if(event->type == MappingNotify)
    updateKeysyms(((XMappingEvent *)event)->first_keycode,
		  ((XMappingEvent *)event)->first_keycode +
		  ((XMappingEvent *)event)->count);
}

static void addKeySym()
{
  KeySym ks;
  int ix;

  if((ks = getKeySym(addKeySym)) != NoSymbol)
    if(currkey) {
      for(ix = MAXKEYSYM; ix; --ix)
	if(currkey->keysym[ix-1] != NoSymbol)
	  break;
      if(ix >= MAXKEYSYM) ix = MAXKEYSYM - 1;
      currkey->keysym[ix] = ks;
      modifyKeySyms(currkey);
      setkeysymName("");
    }
}

static void expandKeySym()
{
  getKeySym(NULL);
}

static void readModmap(char *path)
{
  pid_t pid;

  if(pid = FORK()) {
    if(pid<0)
      perror("fork");
    else
      while(wait(NULL)<0)
	if(errno!=EINTR) {
	  perror("wait");
	  break;
	}
  } else {
    /* child */
    execlp("xmodmap", "xmodmap", path, NULL);
    perror("execlp xmodmap");
    _exit(0);
  }
}

static void writeModmap(char *path)
{
  char *p;
  FILE *f;
  time_t now;
  int mod, c, m, n, mask;
  struct key *k;

  if(f = fopen(path, "w")) {
    time(&now);
    fprintf(f, "! Created with XSetMap v1.0 - %s!\n\n", ctime(&now));
    for(mod=0; mod<8; mod++)
      fprintf(f, "clear %s\n", mod_name[mod]);
    for(k=key, c=0; c<256; c++,k++)
      if(k->w) {
	fprintf(f, "\nkeycode 0x%02x =", c);
	for(m=MAXKEYSYM; m; --m)
	  if(k->keysym[m-1] != NoSymbol)
	    break;
	for(n=0; n<m; n++)
	  if(p = XKeysymToString(k->keysym[n]))
	    fprintf(f, " %s", p);
	  else
	    fprintf(f, " 0x%04lx", (long)k->keysym[n]);
      }
    fprintf(f, "\n");
    for(mask=1, mod=0; mod<8; mask<<=1, mod++)
      for(k=key, c=n=0; c<256; c++,k++)
	if(k->w)
	  if(k->modifiers & mask) {
	    if(!(n++))
	      fprintf(f, "\nadd %s =", mod_name[mod]);
	    if(p = XKeysymToString(k->keysym[0]))
	      fprintf(f, " %s", p);
	    else
	      fprintf(f, " 0x%04lx", (long)k->keysym[0]);
	  }
    fprintf(f, "\n\n!\n");
    fclose(f);
  } else alert("Failed to create \"%s\"", path);
}

static XtActionsRec actions[] = {
  {"remapkbd", (XtActionProc)remapKbd},
  {"remapmod", (XtActionProc)updateModifiers},
  {"addkeysym", (XtActionProc)addKeySym},
  {"expandkeysym", (XtActionProc)expandKeySym},
  {"placecompl", (XtActionProc)placeCompl},
  {"removeparent", (XtActionProc)removeParent},
  {"replacekeysym", (XtActionProc)replaceKeySym},
  {"clearsym", (XtActionProc)clearsymCallback},
};

static char kbdtrans[] = "#override\n\
<Mapping>Keyboard: remapkbd()\n\
<Mapping>Modifier: remapmod()";

static char symtrans[] = "#override\n\
<Btn2Down>,<Btn2Up>:placecompl() Set() replacekeysym()\n\
<Btn3Down>: Set() XawPositionSimpleMenu(keysymedit) XtMenuPopup(keysymedit)";

static char ksnaccel[] = "\
Ctrl<Key>F: forward-character()\n\
Ctrl<Key>B: backward-character()\n\
Ctrl<Key>D: delete-next-character()\n\
Ctrl<Key>A: beginning-of-line()\n\
Ctrl<Key>E: end-of-line()\n\
Ctrl<Key>H: delete-previous-character()\n\
Ctrl<Key>K: kill-to-end-of-line()\n\
Ctrl<Key>W: kill-selection()\n\
Ctrl<Key>Y: insert-selection(SECONDARY)\n\
Ctrl<Key>U: select-all() kill-selection()\n\
Ctrl<Key>space: clearsym()\n\
Ctrl<Key>: no-op(RingBell)\n\
<Key>Right: forward-character()\n\
<Key>Left: backward-character()\n\
<Key>Delete: delete-previous-character()\n\
<Key>BackSpace: delete-previous-character()\n\
<Key>space: no-op()\n\
<Key>Return: placecompl() addkeysym()\n\
<Key>KP_Enter: placecompl() addkeysym()\n\
<Key>Tab: placecompl() expandkeysym()\n\
<Key>: insert-char()";

static char ksntrans[] = "\
<FocusIn>: focus-in()\n\
<FocusOut>: focus-out()\n\
<Btn1Down>: select-start()\n\
<Btn1Motion>: extend-adjust()\n\
<Btn1Up>: extend-end(PRIMARY, CUT_BUFFER0)\n\
<Btn2Down>: insert-selection(PRIMARY, CUT_BUFFER0)\n\
<Btn3Down>: extend-start()\n\
<Btn3Motion>: extend-adjust()\n\
<Btn3Up>: extend-end(PRIMARY, CUT_BUFFER0)";

static char menutrans[] = "#override\
<BtnUp>:MenuPopdown() placecompl() notify() unhighlight()";

static char keytransstring[] = "#override\
<Btn3Down>: set() notify() XawPositionSimpleMenu(alpha) XtMenuPopup(alpha)";

static char aboutText[] = "\
XSetMap 1.0\n\
by Marcus Comstedt\n\
marcus@lysator.liu.se";

void initWidgets(Widget shell, Widget keycode_win)
{
  static char alpha[] = "A\0B\0C\0D\0E\0F\0G\0H\0I\0J\0K\0L\0M\0N\0O\0P\0Q\0R\0S\0T\0U\0V\0W\0X\0Y\0Z\0000\0001\0002\0003\0004\0005\0006\0007\0008\0009\0";
  char *p;
  Widget box, mnu, w, about_win;
  int i;
  
  XtAppAddActions(ctx, actions, XtNumber(actions));
  XawSimpleMenuAddGlobalActions(ctx);

  keytrans = XtParseTranslationTable(keytransstring);

  readreq = createFileRequester(shell, "readfile",
				"Select xmodmap file to read", readModmap);
  writereq = createFileRequester(shell, "writefile",
				"Select xmodmap file to write", writeModmap);
  about_win = XtVaCreatePopupShell("about", transientShellWidgetClass, shell,
				   NULL);
  parent = XtVaCreateManagedWidget("form", formWidgetClass, about_win,
				   NULL);
  w = XtVaCreateManagedWidget("text", asciiTextWidgetClass, parent,
			      XtNstring, aboutText,
			      XtNdisplayCaret, FALSE,
			      XtNuseStringInPlace, TRUE,
			      XtNwidth, 300, XtNheight, 75,
			      NULL);

  XtAddCallback(XtVaCreateManagedWidget("ok", commandWidgetClass, parent,
					XtNfromVert, w, NULL),
		XtNcallback, (XtCallbackProc)popdownCallback, (XtPointer)about_win);

  parent = XtVaCreateManagedWidget("panes", panedWidgetClass, shell,
				   NULL);
  box = XtVaCreateManagedWidget("commands", boxWidgetClass, parent,
				XtNorientation, XtorientHorizontal,
				XtNresizeToPreferred, TRUE,
				NULL);
  kbdwidget = XtVaCreateManagedWidget("keyboard", formWidgetClass, parent,
				      XtNshowGrip, FALSE, XtNallowResize, TRUE,
				      NULL);
  mnu = XtVaCreatePopupShell("alpha", simpleMenuWidgetClass, kbdwidget,
			     NULL);
  for(p = alpha; *p; p+=2)
    XtAddCallback(XtVaCreateManagedWidget(p, smeBSBObjectClass, mnu, NULL),
		  XtNcallback, (XtCallbackProc)alphaCallback, (XtPointer)(long)*p);

  modbox = XtVaCreateManagedWidget("modifiers", formWidgetClass, parent,
				   XtNshowGrip, FALSE, XtNskipAdjust, TRUE,
				   NULL);
  for(i=0; i<8; i++)
    XtAddCallback(modwidget[i] =
	  XtVaCreateManagedWidget(mod_name[i], toggleWidgetClass, modbox,
				  XtNfromHoriz, (i? modwidget[i-1] : NULL),
				  NULL),
		  XtNcallback, (XtCallbackProc)changeMod, (XtPointer)i);

  symbox = XtVaCreateManagedWidget("keysympanes", panedWidgetClass, parent,
				   XtNshowGrip, FALSE,
				   XtNorientation, XtorientHorizontal,
				   XtNskipAdjust, TRUE,
				   NULL);
  XtVaCreateManagedWidget("label", labelWidgetClass, symbox,
			  XtNlabel, "Keysyms:", XtNresizeToPreferred, TRUE,
			  NULL);
  symwidget = XtVaCreateManagedWidget("keysyms", listWidgetClass, symbox,
				      XtNdefaultColumns, MAXKEYSYM,
				      XtNforceColumns, TRUE,
				      XtNshowGrip, FALSE,
				      NULL);
  XtAddCallback(symwidget, XtNcallback, (XtCallbackProc)editsymCallback, NULL);

  keysymname =
    XtVaCreateManagedWidget("keysymname", asciiTextWidgetClass, parent,
			    XtNshowGrip, FALSE, XtNskipAdjust, TRUE,
			    XtNeditType, XawtextEdit,
			    XtNtranslations, XtParseTranslationTable(ksntrans),
			    XtNaccelerators, XtParseAcceleratorTable(ksnaccel),
			    NULL);

  XtAddCallback(XtVaCreateManagedWidget("quit", commandWidgetClass, box,
					NULL),
		XtNcallback, (XtCallbackProc)quitOut, NULL);
  XtAddCallback(XtVaCreateManagedWidget("read", commandWidgetClass, box,
					NULL),
		XtNcallback, (XtCallbackProc)popupCallback, (XtPointer)readreq);
  XtAddCallback(XtVaCreateManagedWidget("write", commandWidgetClass, box,
					NULL),
		XtNcallback, (XtCallbackProc)popupCallback, (XtPointer)writereq);
  if(!makeKbdMenu(w=XtVaCreateManagedWidget("keyboard", menuButtonWidgetClass,
					    box, NULL)))
    XtSetSensitive(w, False);
  XtAddCallback(XtVaCreateManagedWidget("custom", commandWidgetClass, box,
					NULL),
		XtNcallback, (XtCallbackProc)popupCallback, (XtPointer)keycode_win);
  XtAddCallback(XtVaCreateManagedWidget("about", commandWidgetClass, box,
					NULL),
		XtNcallback, (XtCallbackProc)popupCallback, (XtPointer)about_win);

  mnu = XtCreatePopupShell("keysymedit", simpleMenuWidgetClass, symwidget,
			   NULL, 0);
  XtAddCallback(XtVaCreateManagedWidget("insert", smeBSBObjectClass, mnu,
					NULL),
		XtNcallback, (XtCallbackProc)insertsymCallback, NULL);
  XtAddCallback(XtVaCreateManagedWidget("remove", smeBSBObjectClass, mnu,
					NULL),
		XtNcallback, (XtCallbackProc)removesymCallback, NULL);
  XtAddCallback(XtVaCreateManagedWidget("clear", smeBSBObjectClass, mnu,
					NULL),
		XtNcallback, (XtCallbackProc)clearsymCallback, NULL);

  XtOverrideTranslations(kbdwidget, XtParseTranslationTable(kbdtrans));
  XtOverrideTranslations(symwidget, XtParseTranslationTable(symtrans));
  XtOverrideTranslations(mnu, XtParseTranslationTable(menutrans));

  XtInstallAccelerators(parent, keysymname);

  selectKc(NULL, NULL);
}
