#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/Xaw/Label.h>
#include <stdio.h>
#include <stdlib.h>

#include "xsetmap.h"

Widget topLevel;
XtAppContext ctx;
struct key key[256];

struct AppData appldata;

static String fb_res[] = {
  "*layout.Width: 950",
  "*layout.Height: 300",
  "*panes.keyboard.Width: 1",
  "*panes.keyboard.Height: 1",
  "*files.Height: 400",
  NULL };

static XrmOptionDescRec options[] = {
  {"-keyboard",  ".keyboard",  XrmoptionSepArg, NULL},
  {"-kbd",       ".keyboard",  XrmoptionSepArg, NULL},
  {"-blocksize", ".blockSize", XrmoptionSepArg, NULL},
};

static XtResource app_res[] = {
  {
    "keyboard",
    "Keyboard",
    XtRString,
    sizeof(char *),
    XtOffset(struct AppData *, def_keyboard),
    XtRString,
    NULL
  },
  {
    "blockSize",
    "BlockSize",
    XtRInt,
    sizeof(int),
    XtOffset(struct AppData *, blocksize),
    XtRImmediate,
    (caddr_t) 10,
  }
};

void clearKeys()
{
  int i;

  for(i=0; i<256; i++) {
    if(key[i].widget2) {
      XtDestroyWidget(key[i].widget2);
      key[i].widget2 = NULL;
    }
    key[i].w=0;
    key[i].keycode=i;
  }
}

static void clearSyms()
{
  int i,j;

  for(i=0; i<255; i++) {
    for(j=0; j<MAXKEYSYM; j++)
      key[i].keysym[j]=NoSymbol;
    key[i].modifiers = 0;
  }
}

static void initKeys()
{
  memset(key, 0, sizeof(key));
  clearKeys();
  clearSyms();
}

void main(unsigned int argc, char *argv[])
{
  Arg arg;

  XtSetArg(arg, XtNallowShellResize, True);
  topLevel = XtAppInitialize(&ctx, "XSetmap", options, XtNumber(options),
			     &argc, argv, fb_res, &arg, 1);
  XtGetApplicationResources(topLevel, (XtPointer)&appldata, app_res,
			    XtNumber(app_res), NULL, 0);
  if(argc > 1) {
    fprintf(stderr, "%s: unable to parse argument: %s\n", argv[0], argv[1]);
    exit(10);
  }
  initKeys();
  updateKeysyms(0,256);
  updateModifiers();
  initWidgets(topLevel, createKeyCodeWin(topLevel));
  if(appldata.def_keyboard)
    if(loadKeyboard(appldata.def_keyboard))
      applyLayout();
  XtRealizeWidget(topLevel);
  XtAppMainLoop(ctx);
  exit(0);
  /*NOTREACHED*/
}
